# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import uuid
from typing import Any, Callable, List, Union

from geodataflow.geoext.dataset import GdalDataset
from geodataflow.geoext.gdalenv import GdalEnv

# Some predefined Satellite band names and indexes.
EO_BAND_NAMES = dict([
    ('S2_MSI_L2A', ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B11', 'B12',
                    'AOT', 'WVP', 'SCL'])
])


class EOGdalDataset(GdalDataset):
    """
    Wrapper of a GDAL Dataset reading EO/STAC Products (Raster Datasets).
    """
    @staticmethod
    def calculate_gdal_path(dataset_path: str, gdal_env: GdalEnv) -> str:
        """
        Returns the GDAL FileSystem path to access to the specified resource.
        """
        if dataset_path.startswith('s3://'):
            return '/vsis3/' + dataset_path[5:]
        if dataset_path.startswith('gs://'):
            return '/vsigs/' + dataset_path[5:]

        if dataset_path.startswith('https://'):
            #
            if all(f in dataset_path for f in ['.s3.', '.amazonaws.com/']):
                temp_ = dataset_path[8:].split('/')
                temp_ = '/'.join([temp_[0].split('.')[0]] + temp_[1:])
                return '/vsis3/' + temp_

            gdal = gdal_env.gdal()
            return '/vsicurl/' + gdal.GetSignedURL(dataset_path)

        return dataset_path

    @staticmethod
    def open(assets_collection: Union[str, List[str]],
             bands: List[str],
             gdal_env: GdalEnv,
             custom_dataset_func: Callable[[GdalDataset, Any], GdalDataset] = None,
             custom_args: Any = None) -> GdalDataset:
        """
        Open the specified EO Product path, it can be a mosaic-list of paths.
        """
        assets_collection = \
            assets_collection if isinstance(assets_collection, list) else [assets_collection]

        if gdal_env is None:
            gdal_env = GdalEnv.default()

        if not bands:
            bands = EO_BAND_NAMES.get('S2_MSI_L2A')

        # Read each EO Product as mosaic of GDAL Datasets.
        datasets = list()
        gdal_config = gdal_env.gdal_config()

        for assets in assets_collection:
            raster_files = [
                EOGdalDataset.calculate_gdal_path(assets[band]["href"], gdal_env) for band in bands
            ]
            virtual_name = str(uuid.uuid1()).replace('-', '')
            virtual_file = os.path.join(gdal_env.temp_data_path(), 'temp_S3_{}.vrt'.format(virtual_name))

            command_line = \
                'gdalbuildvrt {} -separate "{}" {}' \
                .format(gdal_config, virtual_file, ' '.join(raster_files))

            gdal_env.run(command_line)
            datasets.append(GdalDataset(virtual_file, gdal_env))

        # Apply custom transform to input Datasets?
        if custom_dataset_func:
            custom_args = custom_args.copy()
            custom_args['datasets'] = datasets
            custom_args['gdal_env'] = gdal_env
            datasets = [custom_dataset_func(dataset, custom_args) for dataset in datasets]

        # Create output, a mosaic of GDAL datasets.
        dataset = GdalDataset.mosaic_of_datasets(datasets)
        return dataset



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

# TQDM Python module installed?, therefore we can define a Callback Function.
_TQDM_STATE_KEY_ = 0
_TQDM_STATE_REF_ = None


class ProgressProcessingStore(object):
    """
    Shows the progress of a Pipeline Job walking on a DataStore.
    """
    def __init__(self, processing_args=None):
        self.enabled = True

        # Available UI percentage progress?
        if processing_args:
            self.enabled = hasattr(processing_args, 'ui_mode') and \
                bool(getattr(processing_args, 'ui_mode', False))

        self._tqdm = ProgressProcessingStore.validate_tqdm_ref() if self.enabled else None
        self.enabled = self._tqdm is not None
        self._progress_bar = None

    @staticmethod
    def validate_tqdm_ref():
        """
        Check if TQDM module is installed.
        """
        global _TQDM_STATE_KEY_
        global _TQDM_STATE_REF_

        if _TQDM_STATE_KEY_ == 0:
            try:
                from tqdm import tqdm
                _TQDM_STATE_KEY_ = 1
                _TQDM_STATE_REF_ = tqdm
            except Exception:
                import logging
                logging.warning('TQDM module not installed. Maybe Progress UI capabilities be unavailable.')
                _TQDM_STATE_KEY_ = 2
                _TQDM_STATE_REF_ = None

        return _TQDM_STATE_REF_

    def initialized(self):
        """
        Returns whether the Progress UI was already initialized.
        """
        return self._progress_bar is not None

    def initialize(self, item_count: int) -> bool:
        """
        Initializes the Progress UI to the specified count.
        """
        if not self.enabled or item_count <= 0:
            self._progress_bar = None
            return False
        elif self._progress_bar is not None:
            self._progress_bar.reset(item_count)
            return True
        else:
            self._progress_bar = self._tqdm(range(item_count), ncols=80)
            return True

    def update(self) -> None:
        """
        Updates the percentage progress of current Job.
        """
        if self._progress_bar:
            self._progress_bar.update()

        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import importlib
import requests
from typing import Dict, List
from shapely.geometry import mapping as shapely_mapping, shape as shapely_shape
from shapely.geometry.polygon import Polygon
from geodataflow.core.common import DateUtils
from geodataflow.pipeline.basictypes import AbstractFilter

# List of available GEE Datasets.
GEE_CATALOG_URL = \
    'https://raw.githubusercontent.com/samapriya/Earth-Engine-Datasets-List/master/gee_catalog.json'


def search_gee_datasets() -> List[Dict]:
    """
    Discover list of available GEE Datasets.
    """
    try:
        respose = requests.get(GEE_CATALOG_URL)
        catalog_list = respose.json()
        datasets = [dataset for dataset in catalog_list if dataset['type'] == 'image_collection']
        datasets = sorted(datasets, key=lambda obj: obj['title'], reverse=False)
        return datasets
    except Exception:
        return []


class GEEProductCatalog(AbstractFilter):
    """
    The Filter extracts Metadata from GEE Collections via spatial & alphanumeric filters.
    """
    GEE_CATALOG_DATASET = search_gee_datasets() if importlib.util.find_spec('ee') else []

    def __init__(self):
        AbstractFilter.__init__(self)
        self._ee_api = None
        self.dataset = 'COPERNICUS/S2_SR_HARMONIZED'
        self.startDate = None
        self.endDate = None
        self.closestToDate = None
        self.windowDate = None
        self.filter = None
        self.preserveInputCrs = True

    @staticmethod
    def is_available() -> bool:
        """
        Indicates that this Module is available for use, some modules may depend
        on the availability of other third-party packages.
        """
        ee_api_spec = importlib.util.find_spec('ee')
        return ee_api_spec is not None

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'EarthEngine Catalog'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'EO STAC Imagery'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Extracts Metadata from GEE Collections via spatial & alphanumeric filters.'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        dataset_labels = []
        dataset_keys = []

        for dataset in GEEProductCatalog.GEE_CATALOG_DATASET:
            dataset_keys.append(dataset['id'])
            dataset_labels.append(dataset['title'])

        return {
            'dataset': {
                'description': 'Earth Engine Dataset from which to fetch data.',
                'dataType': 'string',
                'default': 'COPERNICUS/S2_SR_HARMONIZED',
                'options': dataset_keys,
                'labels': dataset_labels
            },
            'startDate': {
                'description': 'Start date of images to fetch (Optional). "$TODAY()" is supported.',
                'dataType': 'date'
            },
            'endDate': {
                'description': 'End date of images to fetch (Optional). "$TODAY()" is supported.',
                'dataType': 'date'
            },
            'closestToDate': {
                'description':
                    'Select only those images which Date is the closest to the specified (Optional).',
                'dataType': 'date'
            },
            'windowDate': {
                'description':
                    'Days around "closestToDate" when "startDate" and "endDate" are not specified.',
                'dataType': 'int',
                'default': 5
            },
            'filter': {
                'description': 'Attribute filter string of images to fetch (Optional).',
                'dataType': 'filter'
            },
            'preserveInputCrs': {
                'description': 'Preserve input CRS, otherwise Geometries are transformed to "EPSG:4326".',
                'dataType': 'bool',
                'default': True
            }
        }

    def starting_run(self, schema_def, pipeline, processing_args, fetch_fields: bool = True):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.core.schemadef import DataType, GeometryType, FieldDef
        from geodataflow.geoext.commonutils import GeometryUtils

        envelope = schema_def.envelope if schema_def.envelope else [0, -90, 360, 90]
        geometry = GeometryUtils.create_geometry_from_bbox(envelope[0], envelope[1], envelope[2], envelope[3])

        # Fetch schema from first GEE Image.
        field_dict = {f.name: f for f in schema_def.fields}
        new_fields = [
            FieldDef(name='productType', data_type=DataType.String),
            FieldDef(name='productDate', data_type=DataType.String)
        ]
        if fetch_fields:
            ee = self.ee_initialize()

            ee_dataset = ee.ImageCollection(self.dataset)
            ee_image = ee_dataset.first()
            ee_properties = self.ee_object_props(ee_image)
            properties = ee_properties.getInfo()
            #
            for name, value in properties.items():
                if not field_dict.get(name) and \
                   not name.startswith('system:') and \
                   not isinstance(value, list) and \
                   not isinstance(value, dict):
                    new_fields.append(FieldDef(name=name, data_type=DataType.to_data_type(value)))

        schema_def = schema_def.clone()
        schema_def.geometryType = GeometryType.Polygon
        schema_def.input_srid = schema_def.srid
        schema_def.input_crs = schema_def.crs
        schema_def.fields += new_fields

        # Redefine CRS when distinct of EPSG:4326?
        if not self.preserveInputCrs and schema_def.srid != 4326 and schema_def.envelope:
            from pyproj import CRS
            from geodataflow.geoext.commonutils import GeometryUtils

            schema_def.srid = 4326
            schema_def.crs = CRS.from_epsg(4326)

            transform_fn = GeometryUtils.create_transform_function(schema_def.input_crs, schema_def.crs)
            geometry = GeometryUtils.create_geometry_from_bbox(*schema_def.envelope)
            geometry = transform_fn(geometry)
            schema_def.envelope = list(geometry.bounds)

        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        schema_def = self.pipeline_args.schema_def

        # Transform output Geometries to input CRS?
        from geodataflow.geoext.commonutils import GeometryUtils
        inv_transform_fn = \
            GeometryUtils.create_transform_function(4326, schema_def.input_crs) \
            if self.preserveInputCrs else None

        ee = self.ee_initialize()

        def ee_iter_function(ee_image, new_list):
            """
            Iterating ee.Image's of one ee.ImageCollection().
            """
            new_list = ee.List(new_list)
            ee_image_props = self.ee_object_props(ee_image)
            return ee.List(new_list.add(ee_image_props))

        # Query & fetch GEE imagery from Google Earth Engine...
        index = 0
        for feature in feature_store:
            geometry = feature.geometry

            ee_dataset = self.ee_fetch_dataset(geometry, False)
            ee_images = ee_dataset.iterate(ee_iter_function, ee.List([]))

            for image_props in ee_images.getInfo():
                properties = feature.properties.copy()

                properties['productType'] = self.dataset
                properties['productDate'] = image_props['IMAGE_DATE']

                for name, value in image_props.items():
                    if not name.startswith('system:') and \
                       not isinstance(value, list) and not isinstance(value, dict):
                        properties[name] = value

                g = image_props.get('system:footprint')
                if g:
                    g = shapely_shape(g)
                    g = Polygon(g) if g.geom_type == 'LinearRing' else g
                    g = g.with_srid(4326)
                    g = inv_transform_fn(g) if inv_transform_fn else g
                else:
                    g = geometry

                product = type('Feature', (object,), {
                    'type': 'Feature',
                    'fid': index,
                    'properties': properties,
                    'geometry': g
                })
                yield product
                index += 1

        pass

    def ee_initialize(self):
        """
        Returns a GEE environemt.
        """
        if self._ee_api is None:
            import ee

            # Authenticates and initializes Earth Engine.
            try:
                ee.Initialize()
            except Exception:
                # Get credentials:
                # https://developers.google.com/earth-engine/guides/python_install-conda#get_credentials
                ee.Authenticate()
                ee.Initialize()

            self._ee_api = ee
            #
        return self._ee_api

    def ee_object_props(self, ee_object):
        """
        Get the properties of the specified GEE object.
        """
        ee = self.ee_initialize()

        if isinstance(ee_object, ee.Image):
            image = ee_object

            prop_names = image.propertyNames()
            values = prop_names.map(lambda prop_name: image.get(prop_name))
            dictionary = ee.Dictionary.fromLists(prop_names, values)

            image_date = ee.Date(image.get('system:time_start')).format('yyyy-MM-dd')
            dictionary = dictionary.set('IMAGE_DATE', image_date)

            band_names = image.bandNames()
            scales = band_names.map(lambda b: image.select([b]).projection().nominalScale())
            scale = ee.Algorithms.If(scales.distinct().size().gt(1),
                                     ee.Dictionary.fromLists(ee.List(band_names), scales),
                                     scales.get(0))

            dictionary = dictionary.set('system:band_scales', scale)
            return dictionary

        if isinstance(ee_object, ee.Feature):
            feature = ee_object

            prop_names = feature.propertyNames()
            values = prop_names.map(lambda prop_name: feature.get(prop_name))

            return ee.Dictionary.fromLists(prop_names, values)

        if isinstance(ee_object, ee.ImageCollection):
            dataset = ee_object

            geometry = dataset.geometry()
            count = dataset.size()
            keys = dataset.aggregate_array('system:id')

            # Get list of dates of current Collection.
            dates = dataset \
                .map(lambda img: ee.Feature(None, {'date': img.date().format('yyyy-MM-dd')})) \
                .distinct('date') \
                .aggregate_array('date')

            prop_names = ee.List(['geometry', 'dates', 'keys', 'count'])
            values = ee.List([geometry, dates, keys, count])

            return ee.Dictionary.fromLists(prop_names, values)

        return ee.Dictionary({})

    def ee_fetch_dataset(self, geometry, clip_images: bool = True):
        """
        Fetch the GEE ImageCollection according to current settings.
        """
        ee = self.ee_initialize()

        start_date, final_date = \
            DateUtils.parse_date_range(self.startDate, self.endDate, self.closestToDate, self.windowDate)

        start_date = start_date.strftime('%Y-%m-%d')
        final_date = final_date.strftime('%Y-%m-%d')

        from geodataflow.geoext.commonutils import GeometryUtils
        transform_fn = GeometryUtils.create_transform_function(geometry.get_srid(), 4326)
        geometry = transform_fn(geometry)
        geo_json = shapely_mapping(geometry)

        # Define settings for querying...
        ee_dataset = ee.ImageCollection(self.dataset)
        ee_dataset = ee_dataset.filterBounds(geo_json).filterDate(start_date, final_date)

        if self.filter:
            ee_dataset = ee_dataset.filter(self.filter)

        if clip_images:
            ee_dataset = ee_dataset.map(lambda image: image.clip(geo_json))

        ee_dataset = ee_dataset.sort('system:time_start')
        return ee_dataset



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import sys
import threading
from typing import Dict
from geodataflow.core.processingargs import ProcessingArgs

# Default GDAL/OGR environment.
_DEFAULT_GDAL_LOCK = threading.RLock()
_DEFAULT_GDAL_ENVS: Dict[int, "GdalEnv"] = dict()


class GdalEnv(ProcessingArgs):
    """
    Very simple abstraction of GDAL/OGR environment.

    NOTE:
        Fiona provides better abstraction but for now it is not necessary
        to add that dependency.

    Args:
        config_options: GDAL/OGR config options.
        temp_path: Path where temporary files are saved.
    """
    def __init__(self, config_options: Dict[str, str] = None, temp_path: str = None):
        ProcessingArgs.__init__(self, temp_path=temp_path)
        self._config_options = config_options or {}
        self._gdal = None
        self._ogr = None

    def dispose(self) -> None:
        """
        Dispose all resources including temporary data folder.
        """
        global _DEFAULT_GDAL_LOCK
        global _DEFAULT_GDAL_ENVS

        with _DEFAULT_GDAL_LOCK:
            thread_id = threading.get_ident()
            gdal_env = _DEFAULT_GDAL_ENVS.get(thread_id)

            if gdal_env is not None:
                del _DEFAULT_GDAL_ENVS[thread_id]
                gdal_env.dispose()

        ProcessingArgs.dispose(self)

    @staticmethod
    def default() -> "GdalEnv":
        """
        Default GDAL/OGR environment.
        """
        global _DEFAULT_GDAL_LOCK
        global _DEFAULT_GDAL_ENVS

        with _DEFAULT_GDAL_LOCK:
            thread_id = threading.get_ident()
            gdal_env = _DEFAULT_GDAL_ENVS.get(thread_id)

            if gdal_env is None:
                gdal_env = GdalEnv(config_options=GdalEnv.default_options(), temp_path='')
                _DEFAULT_GDAL_ENVS[thread_id] = gdal_env

            return gdal_env

    @staticmethod
    def default_options() -> Dict[str, str]:
        """
        Default GDAL/OGR options for GdalEnv instances.
        """
        return {
            'CPL_VSIL_CURL_ALLOWED_EXTENSIONS': '.tif,.tiff,.vrt,.ovr',
            'CPL_CURL_VERBOSE': 'NO',
            'CPL_DEBUG': 'NO',
            'CPL_VSIL_CURL_USE_CACHE': 'TRUE',
            'CPL_VSIL_CURL_CACHE_SIZE': '64000000',
            'GDAL_DISABLE_READDIR_ON_OPEN': 'FALSE',
            'GDAL_NUM_THREADS': 'ALL_CPUS',
            'GDAL_CACHEMAX': '128',
            'GDAL_TIFF_INTERNAL_MASK': 'YES',
            'VSI_CACHE': 'TRUE',
            'VSI_CACHE_SIZE': '32000000'
        }

    @staticmethod
    def gdal_scripts_path() -> str:
        """
        Returns the root folder where GDAL utility scripts (e.g. gdal_calc.py) are installed.
        """
        is_windows = sys.platform.startswith('win')
        return sys.prefix + '/scripts' if is_windows else '/usr/bin'

    def config_options(self) -> Dict[str, str]:
        """
        Returns configured GDAL/OGR options of this GdalEnv.
        """
        return self._config_options

    def gdal_config(self) -> str:
        """
        Returns configured GDAL/OGR options as a String.
        """
        text_config = ''

        for k, v in self._config_options.items():
            text_config += ' --config {} {}'.format(k, v)

        return text_config

    def gdal(self):
        """
        Returns the GDAL module managed by this GdalEnv.
        """
        if self._gdal is None:
            from osgeo import gdal
            gdal.UseExceptions()

            for k, v in self._config_options.items():
                gdal.SetConfigOption(k, v)

            self._gdal = gdal

        return self._gdal

    def ogr(self):
        """
        Returns the OGR module managed by this GdalEnv.
        """
        if self._ogr is None:
            from osgeo import ogr
            ogr.UseExceptions()
            self._ogr = ogr

        return self._ogr



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import logging
from typing import Iterable, List

from geodataflow.core.schemadef import FieldDef, SchemaDef
from geodataflow.geoext.commonutils import GdalUtils

from pyproj.crs import CRS
from osgeo import ogr
from osgeo import osr
ogr.UseExceptions()
osr.UseExceptions()


class OgrFeatureLayer(object):
    """
    Wrapper class of the OGR FeatureLayer object.
    """
    def __init__(self, data_source, feature_layer):
        self._data_source = data_source
        self._layer = feature_layer

    def __del__(self):
        self._data_source = None
        self._layer = None

    def data_source(self):
        """
        Returns the OGR DataSource related to this Feature Layer.
        """
        return self._data_source

    def layer(self):
        """
        Returns the OGR Layer managed.
        """
        return self._layer

    def features(self) -> Iterable:
        """
        Enumerate Features of this Feature Layer.
        """
        feature_layer = self._layer
        feature = feature_layer.GetNextFeature()
        #
        # NOTE:
        # We can't use "for feature in feature_layer:" -> https://github.com/nextgis/pygdal/issues/31
        #
        while feature is not None:
            yield feature
            feature = feature_layer.GetNextFeature()

        pass

    @staticmethod
    def create_spatial_reference(schema_def: SchemaDef) -> "osr.SpatialReference":
        """
        Creates an `osr.SpatialReference` from the specified SchemaDef.
        """
        spatial_ref = osr.SpatialReference()

        if hasattr(spatial_ref, 'SetAxisMappingStrategy'):
            spatial_ref.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

        if schema_def.srid > 0:
            spatial_ref.ImportFromEPSG(schema_def.srid)
        elif schema_def.crs:
            crs = schema_def.crs

            spatial_ref.ImportFromEPSG(crs.to_epsg()) \
                if crs.to_epsg() else spatial_ref.ImportFromWkt(crs.to_wkt())
        else:
            spatial_ref.ImportFromEPSG(4326)

        return spatial_ref

    @staticmethod
    def create_layer(connection_string: str,
                     layer_name: str,
                     schema_def: SchemaDef,
                     format_options: List[str] = []) -> "OgrFeatureLayer":
        """
        Creates a new OGR FeatureLayer from the specified SchemaDef.
        """
        driver = GdalUtils.get_ogr_driver(connection_string)
        geometry_type = schema_def.geometryType
        spatial_ref = OgrFeatureLayer.create_spatial_reference(schema_def)

        # Remove old existing FeatureStore.
        try:
            if os.path.exists(connection_string):
                driver.DeleteDataSource(connection_string)

        except Exception as e:
            logging.warning(
                'Fail deleting old FeatureStore of "{}", Cause: "{}".'.format(connection_string, e))

        # Create FeatureLayer.
        driver_name = driver.GetName()
        feature_store = driver.CreateDataSource(connection_string, options=format_options)
        feature_layer = feature_store.CreateLayer(layer_name, srs=spatial_ref, geom_type=geometry_type)

        for field_def in schema_def.fields:
            field = ogr.FieldDefn(field_def.name, field_def.type)
            if field_def.width:
                field.SetWidth(field_def.width)
            if field_def.precision:
                field.SetPrecision(field_def.precision)
            if field_def.nullable:
                field.SetNullable(1 if field_def.nullable else 0)
            if field_def.defaultValue:
                field.SetDefault(field_def.defaultValue)

            feature_layer.CreateField(field)

        # Fix current saved PRJ file, the 'ESRI Shapefile' driver saves an invalid WKT spec.
        if driver_name == 'ESRI Shapefile':
            shp_file = connection_string
            prj_file = os.path.join(os.path.dirname(shp_file),
                                    os.path.splitext(os.path.basename(shp_file))[0]+'.prj')

            feature_layer.SyncToDisk()
            feature_store.SyncToDisk()
            feature_layer = None
            feature_store = None

            if spatial_ref is not None:
                with open(prj_file, 'w') as file:
                    file.write(spatial_ref.ExportToWkt())

            feature_store = driver.CreateDataSource(os.path.dirname(shp_file))
            layer_name = GdalUtils.get_layer_name(shp_file)
            feature_layer = feature_store.GetLayerByName(layer_name)

        return OgrFeatureLayer(feature_store, feature_layer)

    def get_schema_def(self) -> SchemaDef:
        """
        Returns the schema definition of this FeatureLayer.
        """
        feature_layer = self._layer
        spatial_ref = feature_layer.GetSpatialRef()
        table_def = feature_layer.schema

        if spatial_ref:
            srid = GdalUtils.get_spatial_srid(spatial_ref)
            crs = CRS.from_epsg(srid) if srid else CRS.from_wkt(spatial_ref.ExportToWkt())
        else:
            srid = 0
            crs = None

        envelope = feature_layer.GetExtent()
        envelope = [envelope[0], envelope[2], envelope[1], envelope[3]]

        fields_def = [
            FieldDef(name=fd.GetName(), data_type=fd.GetType(),
                     precision=fd.GetPrecision(), width=fd.GetWidth(),
                     is_nullable=True if fd.IsNullable() else False, default_value=fd.GetDefault(),
                     typeName=fd.GetTypeName())

            for fd in table_def
        ]
        return SchemaDef(type='FeatureLayer',
                         name=feature_layer.GetName(),
                         srid=srid,
                         crs=crs,
                         geometryType=feature_layer.GetGeomType(),
                         envelope=envelope,
                         fields=fields_def)



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import uuid
import requests
import importlib
from zipfile import ZipFile
from typing import Dict, Iterable
from geodataflow.pipeline.filters.GEEProductCatalog import GEEProductCatalog


class GEEProductDataset(GEEProductCatalog):
    """
    The Filter extracts Datasets from GEE Collections via spatial & alphanumeric filters.
    """
    def __init__(self):
        GEEProductCatalog.__init__(self)
        self.configVars = ''
        self.bands = []
        self.groupByDate = True
        self.clipByAreaOfInterest = True

    @staticmethod
    def is_available() -> bool:
        """
        Indicates that this Module is available for use, some modules may depend
        on the availability of other third-party packages.
        """
        ee_api_spec = importlib.util.find_spec('ee')
        return ee_api_spec is not None

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'EarthEngine Dataset'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Extracts Datasets from GEE Collections via spatial & alphanumeric filters.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'EO STAC Imagery'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        the_params = GEEProductCatalog.params(self)
        the_params['configVars'] = {
            'description':
                'Environment variables separated by commas. Commonly used to configure credentials.',
            'dataType': 'string',
            'default': ''
        }
        the_params['bands'] = {
            'description': 'List of Bands to fetch, or a string separated by commas. Empty means fetch all.',
            'dataType': 'string',
            'default': 'B4,B3,B2,B8',
            'placeHolder': 'B4,B3,B2,B8'
        }
        the_params['groupByDate'] = {
            'description': 'Group EO Products by Date.',
            'dataType': 'bool',
            'default': True
        }
        the_params['clipByAreaOfInterest'] = {
            'description': 'Clip EO Products by geometry of input AOI.',
            'dataType': 'bool',
            'default': True
        }
        return the_params

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.core.schemadef import GeometryType, DataType, FieldDef
        from geodataflow.geoext.dataset import DATASET_DEFAULT_SCHEMA_DEF

        new_fields = [
            FieldDef("productType", DataType.String),
            FieldDef("productDate", DataType.String)
        ]

        schema_def = GEEProductCatalog.starting_run(self, schema_def, pipeline, processing_args, False)
        schema_def = schema_def.clone()
        schema_def.geometryType = GeometryType.Polygon
        schema_def.fields = FieldDef.concat(DATASET_DEFAULT_SCHEMA_DEF, new_fields)
        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        schema_def = self.pipeline_args.schema_def

        from geodataflow.geoext.gdalenv import GdalEnv
        from geodataflow.geoext.dataset import GdalDataset

        ee_config_options = GdalEnv.default_options()
        if self.configVars:
            for item in self.configVars.split(','):
                pair = item.split('=')
                ee_config_options[pair[0].strip()] = pair[1].strip() if len(pair) > 1 else ''

        # Query & fetch GEE imagery from Google Earth Engine...
        with GdalEnv(config_options=ee_config_options, temp_path=processing_args.temp_data_path()) as ee_env:
            #
            for feature in feature_store:
                geometry = feature.geometry

                ee_dataset = self.ee_fetch_dataset(geometry, self.clipByAreaOfInterest)
                ee_dataset_props = self.ee_object_props(ee_dataset)
                ee_first_image = ee_dataset.first()
                ee_image_props = self.ee_object_props(ee_first_image)

                image_props = ee_image_props.getInfo()
                band_scales = image_props['system:band_scales']

                # For each Date, get the mosaic the subset of GEE images.
                for product_date, ee_subset in self.ee_enumerate_subsets(ee_dataset, ee_dataset_props, ee_image_props):
                    if self.bands:
                        bands = self.bands.replace(' ', '').split(',') if isinstance(self.bands, str) else self.bands
                        ee_subset = ee_subset.select(bands)
                        factor_scale = band_scales[bands[0]]
                    else:
                        bands = list(band_scales.keys())
                        factor_scale = band_scales[bands[0]]

                    # Query mosaic trying several scales upto it fits Google Erath Engine requirements.
                    temp_name = 'temp_GEE_' + str(uuid.uuid4()).replace('-', '')
                    factor_count = 10
                    url = None
                    while factor_count > 0:
                        try:
                            params = {
                                'name': temp_name,
                                'filePerBand': False,
                                'scale': factor_scale,
                                'region': ee_subset.geometry(),
                                'crs': 'EPSG:' + str(schema_def.srid),
                                'fileFormat': 'GEO_TIFF',
                                'formatOptions': 'TILED=YES'
                            }
                            factor_count -= 1
                            url = ee_subset.getDownloadURL(params)
                            break
                        except Exception:
                            factor_scale *= 2
                            url = None

                    if not url:
                        raise Exception('Request does not fit GEE requirements, query smaller areas.')

                    # Fetch & save request to Geotiff (GEE outputs ZIP files).
                    response = requests.get(url, stream=True)
                    if response.status_code != 200:
                        raise Exception('An error occurred while downloading a GEE image.')

                    output_file = os.path.join(ee_env.temp_data_path(), temp_name + '.zip')
                    raster_file = os.path.join(ee_env.temp_data_path(), temp_name + '.tif')

                    with open(output_file, 'wb') as fp:
                        for chunk in response.iter_content(chunk_size=1024*1024):
                            if chunk:
                                fp.write(chunk)

                    with ZipFile(output_file, mode='r') as zf:
                        zf.extractall(ee_env.temp_data_path())

                    os.remove(output_file)

                    user_data = {
                        'productType': self.dataset,
                        'productDate': product_date
                    }
                    yield GdalDataset(raster_file, ee_env, user_data, False)
                #

        pass

    def ee_enumerate_subsets(self, ee_dataset, ee_dataset_props, ee_image_props) -> Iterable:
        """
        Enumerate the GEE Image set of the ImageCollection.
        """
        ee = self.ee_initialize()
        dataset_props = ee_dataset_props.getInfo()

        if self.groupByDate:
            for product_date in dataset_props['dates']:
                y, m, d = \
                    int(product_date[0:4]), int(product_date[5:7]), int(product_date[8:10])

                ee_subset = ee_dataset \
                    .filter(ee.Filter.calendarRange(y, y, 'year')) \
                    .filter(ee.Filter.calendarRange(m, m, 'month')) \
                    .filter(ee.Filter.calendarRange(d, d, 'day_of_month'))

                ee_mosaic = ee_subset.mosaic()
                ee_mosaic = ee_mosaic.set(ee_image_props)

                yield product_date, ee_mosaic
        else:
            count = dataset_props['count']

            for i in range(0, count):
                ee_image = ee.Image(ee_dataset.toList(count).get(i))
                ee_product_date = ee.Date(ee_image.get('system:time_start')).format('yyyy-MM-dd')
                yield ee_product_date.getInfo(), ee_image

        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
from typing import Dict
from geodataflow.pipeline.basictypes import AbstractFilter


class FeatureLimit(AbstractFilter):
    """
    The Filter validates that input Geometries do not be greater than a Limit.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.fullAreaLimit = None
        self.areaLimit = None
        self.countLimit = None

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Limits'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Validates that input Geometries do not be greater than a Limit.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Feature'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'fullAreaLimit': {
                'description': 'Maximum area covered by all input Geometries (Optional).',
                'dataType': 'float'
            },
            'areaLimit': {
                'description': 'Maximum area covered by each input Geometry (Optional).',
                'dataType': 'float'
            },
            'countLimit': {
                'description': 'Maximum number of input Geometries (Optional).',
                'dataType': 'int'
            }
        }

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        schema_def = self.pipeline_args.schema_def
        crs = schema_def.crs

        # Input parameters.
        full_area_limit = self.fullAreaLimit if self.fullAreaLimit is not None else 0.0
        area_limit = self.areaLimit if self.areaLimit is not None else 0.0
        full_area = 0.0
        count_limit = self.countLimit if self.countLimit is not None else 0
        feature_count = 0

        logging.info(
            'Verifying limits of input data (Area={}, FullArea={}, Count={})...'
            .format(area_limit, full_area_limit, count_limit))

        from shapely.geometry.polygon import Polygon
        from shapely.geometry.multipolygon import MultiPolygon

        # Do we need to transform input Geometries to ellipsoid WGS84 to calculate Areas in m2?
        if crs and not crs.is_projected:
            import pyproj as pj
            from geodataflow.geoext.commonutils import GeometryUtils

            ellipsoid_ob = pj.Geod(ellps='WGS84')
            transform_fn = GeometryUtils.create_transform_function(crs, pj.CRS.from_epsg(4326))

            for feature in feature_store:
                geometry = feature.geometry

                if count_limit and feature_count >= count_limit:
                    raise Exception(
                        'The number of input Features is greater than maximum allowed ({}).'
                        .format(count_limit))

                if isinstance(geometry, (Polygon, MultiPolygon)):
                    geometry = transform_fn(geometry)

                    area, _ = ellipsoid_ob.geometry_area_perimeter(geometry)
                    area = abs(area)
                    full_area += area

                    if area_limit and area > area_limit:
                        raise Exception(
                            'Area of input Geometry ({}) is greater than maximum allowed ({}).'
                            .format(area_limit, area))

                    if full_area_limit and full_area > full_area_limit:
                        raise Exception(
                            'Area of input Layer ({}) is greater than maximum allowed ({}).'
                            .format(full_area_limit, full_area))

                feature_count += 1
                yield feature
            #
        else:
            for feature in feature_store:
                geometry = feature.geometry

                if count_limit and feature_count >= count_limit:
                    raise Exception(
                        'The number of input Features is greater than maximum allowed ({}).'
                        .format(count_limit))

                if isinstance(geometry, (Polygon, MultiPolygon)):
                    area = geometry.area
                    full_area += area

                    if area_limit and area > area_limit:
                        raise Exception(
                            'Area of input Geometry ({}) is greater than maximum allowed ({}).'
                            .format(area_limit, area))

                    if full_area_limit and full_area > full_area_limit:
                        raise Exception(
                            'Area of input Layer ({}) is greater than maximum allowed ({}).'
                            .format(full_area_limit, full_area))

                feature_count += 1
                yield feature
            #

        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.pipeline.modules import AbstractModule


class AbstractFilter(AbstractModule):
    """
    Abstract Filter that operates on items of Geospatial data.
    """
    def __init__(self):
        AbstractModule.__init__(self)
        self.classType = 'filter'


class AbstractReader(AbstractModule):
    """
    Abstract Module that reads items from a Geospatial DataSource.
    """
    def __init__(self):
        AbstractModule.__init__(self)
        self.classType = 'reader'

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns if this Module supports the specified ConnectionString and named StoreCapability.
        """
        return False


class AbstractWriter(AbstractModule):
    """
    Abstract Module that writes items to a Geospatial DataStore.
    """
    def __init__(self):
        AbstractModule.__init__(self)
        self.classType = 'writer'

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns if this Module supports the specified ConnectionString and named StoreCapability.
        """
        return False



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
from typing import Iterable, List, Union

from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.core.schemadef import SchemaDef, GeometryType
from geodataflow.geoext.commonutils import DataUtils, GdalUtils, GeometryUtils
from geodataflow.geoext.dataset import GdalDataset, DATASET_DEFAULT_SCHEMA_DEF
from geodataflow.geoext.gdalenv import GdalEnv


class GdalRasterStore:
    """
    Store that reads & writes Datasets in a GDAL RasterStore.
    """
    def __init__(self):
        self._datasets: List[GdalDataset] = list()

    def __enter__(self):
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.close()

    def __del__(self):
        """
        Releases resources of this GDAL RasterStore.
        """
        self.close()

    def open(self, connection_string: Union[str, Iterable[str]], count_limit: int = None) -> SchemaDef:
        """
        Opens one GDAL RasterStore for the specified ConnectionString.
        """
        dataset_count = 0
        count_limit = int(count_limit) if count_limit is not None else -1
        schema_def = None

        gdal_env = GdalEnv.default()
        gdal = gdal_env.gdal()

        for item_string in DataUtils.enumerate_single_connection_string(connection_string):
            #
            if not GdalUtils.test_gdal_capability(item_string, StoreCapabilities.READ):
                raise Exception(
                    'The GDAL Driver of "{}" does not support the Access mode "READ".'.format(item_string))

            if count_limit > dataset_count:
                logging.warning(
                    'Early processing break! The "rasterCountLimit" flag was specified (Value={}).'
                    .format(count_limit))

                break

            dataset = gdal.Open(item_string, gdal.GA_ReadOnly)
            dataset = GdalDataset(dataset, gdal_env)

            if schema_def is None:
                crs = GeometryUtils.get_spatial_crs(dataset.get_spatial_srid())

                schema_def = SchemaDef(type='RasterLayer',
                                       name=DataUtils.get_layer_name(item_string),
                                       srid=crs.to_epsg(),
                                       crs=crs,
                                       geometryType=GeometryType.Polygon,
                                       envelope=dataset.get_envelope(),
                                       fields=DATASET_DEFAULT_SCHEMA_DEF.copy())

            self._datasets.append(dataset)
            dataset_count += 1

        return schema_def

    def close(self) -> bool:
        """
        Releases the OGR FeatureLayer set managed by this GDAL RasterStore.
        """
        for i in range(0, len(self._datasets)):
            self._datasets[i] = None

        self._datasets = list()
        return True

    def datasets(self) -> Iterable[GdalDataset]:
        """
        Returns the Datasets set managed by this GDAL RasterStore.
        """
        return self._datasets



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
import json
from typing import Iterable, List, Union

from shapely.wkb import loads as shapely_wkb_loads, dumps as shapely_wkb_dumps
from osgeo import ogr
from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.core.schemadef import SchemaDef
from geodataflow.geoext.commonutils import DataUtils, GdalUtils
from geodataflow.geoext.gdalenv import GdalEnv
from geodataflow.geoext.featurelayer import OgrFeatureLayer
ogr.UseExceptions()


class OgrFeatureStore:
    """
    Store that reads & writes Features with Geometries in one OGR FeatureSource.
    """
    def __init__(self):
        self._layers: List[OgrFeatureLayer] = list()

    def __enter__(self):
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.close()

    def __del__(self):
        """
        Releases resources of this OGR FeatureStore.
        """
        self.close()

    def open(self,
             connection_string: Union[str, Iterable[str]],
             access_mode: str = 'r',
             spatial_filter: str = '',
             attribute_filter: str = '') -> SchemaDef:
        """
        Opens one OGR FeatureStore for the specified ConnectionString.
        """
        mode = 0 if access_mode == 'r' else 1

        for item_string in DataUtils.enumerate_single_connection_string(connection_string):
            is_geojson = \
                DataUtils.represents_json_feature_collection(item_string)

            if is_geojson:
                from uuid import uuid4
                feature_class = 'FeatureCollection-{}'.format(uuid4().hex)

                if isinstance(item_string, dict):
                    item_string = json.dumps(item_string)

                item_string = item_string.replace('\'', '"')  # or OGR library fails!
                mmap_name = '/vsimem/{}.json'.format(feature_class)

                gdal_env = GdalEnv.default()
                gdal = gdal_env.gdal()

                logging.debug('Opening "GeoJSON" from memory string...')

                gdal.FileFromMemBuffer(mmap_name, item_string)
                store = ogr.Open(mmap_name, mode)  # 0 means read-only. 1 means writeable.
                layer = store.GetLayerByName(feature_class)
                gdal.Unlink(mmap_name)

                logging.debug('Ok!')
            else:
                feature_class = DataUtils.get_layer_name(item_string)
                connect_store = GdalUtils.get_ogr_connection_string(item_string)

                logging.debug('Opening "{}"...'.format(item_string))

                store = ogr.Open(connect_store, mode)  # 0 means read-only. 1 means writeable.
                layer = store.GetLayerByName(feature_class)
                if layer is None:
                    layer = store.GetLayerByIndex(0)

                logging.debug('Ok!')

            do_spatial_filter = spatial_filter is not None and len(spatial_filter) > 0

            # Set the spatial and/or alphanumeric filters.
            if attribute_filter:
                layer.SetAttributeFilter(attribute_filter)

            if do_spatial_filter:
                for geom_key in ['RECT(', 'EXTENT(', 'BOUNDS(', 'BBOX(', 'ENV(', 'ENVELOPE(']:
                    if spatial_filter.startswith(geom_key):
                        bbox = [float(val.strip()) for val in spatial_filter[len(geom_key):-1].split(',')]
                        layer.SetSpatialFilterRect(bbox[0], bbox[1], bbox[2], bbox[3])
                        do_spatial_filter = False
                        break
            if do_spatial_filter:
                layer.SetSpatialFilter(spatial_filter)

            self._layers.append(OgrFeatureLayer(store, layer))
        #
        return self._layers[0].get_schema_def()

    def create(self, connection_string: str, schema_def: SchemaDef, format_options: List[str] = []) -> SchemaDef:
        """
        Creates a new OGR FeatureStore for the specified ConnectionString and SchemaDef.
        """
        for item_string in DataUtils.enumerate_single_connection_string(connection_string):
            #
            if not GdalUtils.test_ogr_capability(item_string, StoreCapabilities.CREATE):
                raise Exception('The OGR Driver of "{}" does not support Data creation!'.format(item_string))

            layer_obj = \
                OgrFeatureLayer.create_layer(item_string, DataUtils.get_layer_name(item_string), schema_def, format_options)

            self._layers.append(layer_obj)
            break

        return self._layers[0].get_schema_def()

    def close(self) -> bool:
        """
        Releases the OGR FeatureLayer set managed by this OGR FeatureStore.
        """
        for i in range(0, len(self._layers)):
            self._layers[i] = None

        self._layers = list()
        return True

    def features(self) -> Iterable:
        """
        Enumerates the whole collection of Features managed by this OGR FeatureStore.
        """
        for feature_layer in self._layers:
            schema_def = feature_layer.get_schema_def()
            fields = schema_def.fields

            for feature in feature_layer.features():
                geometry = feature.GetGeometryRef()
                geometry = shapely_wkb_loads(bytes(geometry.ExportToWkb()))
                geometry = geometry.with_srid(schema_def.srid)

                feature = type('Feature', (object,), {
                    'type': 'Feature',
                    'fid': feature.GetFID(),
                    'properties': {
                        fields[i].name: feature.GetField(i) for i in range(feature.GetFieldCount())
                    },
                    'geometry': geometry
                })
                yield feature
            #
        pass

    def write_features(self, features: Iterable, cache_size: int = None) -> Iterable:
        """
        Write the specified collection of Features to this OGR FeatureStore.
        """
        feature_transaction_count = 0
        feature_count = 0
        feature_layer = self._layers[0]

        # Group n features per transaction (default 20000).
        # Increase the value for better performance when writing into DBMS drivers that have transaction
        # support. n can be set to unlimited to load the data into a single transaction.
        # https://github.com/OSGeo/gdal/blob/release/3.2/gdal/swig/python/samples/ogr2ogr.py#L1540
        # https://gdal.org/programs/ogr2ogr.html#cmdoption-ogr2ogr-gt
        n = int(cache_size) if cache_size is not None else 20000

        ogr_layer = feature_layer.layer()
        ogr_schema_def = ogr_layer.GetLayerDefn()
        write_geoms = ogr_schema_def.GetGeomFieldCount() > 0
        ogr_layer.StartTransaction()

        for feature in features:
            geometry = shapely_wkb_dumps(feature.geometry)
            geometry = ogr.CreateGeometryFromWkb(geometry)

            fid = feature.fid if hasattr(feature, 'fid') else feature_count
            ogr_feature = ogr.Feature(ogr_schema_def)
            if fid is not None:
                ogr_feature.SetFID(fid)

            if write_geoms:
                ogr_feature.SetGeometry(geometry)

            for k, v in feature.properties.items():
                i = ogr_feature.GetFieldIndex(k)
                if i != -1:
                    ogr_feature.SetField(i, v)

            ogr_layer.CreateFeature(ogr_feature)

            # Group n features per transaction, therefore divide full transaction in parts.
            if feature_transaction_count > n:
                ogr_layer.CommitTransaction()
                ogr_layer.StartTransaction()
                feature_transaction_count = 0

            feature_transaction_count += 1
            feature_count += 1
            yield feature

        ogr_layer.CommitTransaction()

    def layers(self) -> Iterable[OgrFeatureLayer]:
        """
        Returns the OGR FeatureLayer set managed by this OGR FeatureStore.
        """
        return self._layers



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
from typing import Dict, Iterable, Type
from geodataflow.core.modulemanager import ModuleManager


class ProductDriverApi:
    """
    Defines an interface to implement EO Products Providers.
    """
    def __init__(self):
        pass

    @staticmethod
    def is_available() -> bool:
        """
        Indicates that this Module is available for use, some modules may depend
        on the availability of other third-party packages.
        """
        return True

    def name(self) -> str:
        """
        Returns the Name of the Driver.
        """
        raise NotImplementedError('Interface Class has not to implement any method.')

    def search(self,
               provider: str,
               config: Dict,
               product_type: str,
               area=None,
               date=None,
               **keywords) -> Iterable[object]:
        """
        Search the available EO products with the coordinates of an area, a date interval
        and any other search keywords accepted by the OpenSearch API.
        See:
        https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/FullTextSearch?redirectedfrom=SciHubUserGuide.3FullTextSearch
        """
        raise NotImplementedError('Interface Class has not to implement any method.')


class ProductCatalog:
    """
    Provides methods to search and download EO products (Copernicus Open Access Hub, DIAS, AWS, ...).
    """
    def __init__(self):
        self._modules: Dict[str, Type] = None
        self._drivers: Dict[str, Type] = None

    def available_drivers(self):
        """
        Returns the available list of EO Products Provider drivers.
        """
        if self._drivers:
            return self._drivers

        modules_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'drivers')

        manager = ModuleManager(abstract_types=(ProductDriverApi,))
        self._modules = manager.load_modules(modules_folders=[modules_folder])

        self._drivers = dict()
        for ty in self._modules.values():
            driver_ob = ty()
            self._drivers[driver_ob.name().lower()] = driver_ob

        return self._drivers.values()

    def get_driver(self, driver_name: str):
        """
        Returns the EO Products Provider driver of specified name.
        """
        if not self._drivers:
            _ = self.available_drivers()

        return self._drivers.get(driver_name.lower())

    def search(self,
               driver: str,
               provider: str,
               config: Dict,
               product_type: str,
               area=None,
               date=None,
               **keywords) -> Iterable[object]:
        """
        Search the available EO products with the coordinates of an area, a date interval
        and any other search keywords accepted by the OpenSearch API.
        See:
        https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/FullTextSearch?redirectedfrom=SciHubUserGuide.3FullTextSearch
        """
        driver_ob = self.get_driver(driver_name=driver)

        for product_ob in driver_ob.search(
                provider=provider, config=config, product_type=product_type, area=area, date=date, **keywords
                ):
            product_ob = ProductCatalog._normalize_product(product_type, product_ob)
            yield product_ob

        pass

    @staticmethod
    def _normalize_product(product_type: str, product: object) -> object:
        """
        Normalize properties of specified EOProduct from different EO Providers.
        """
        product.geometry = product.geometry.with_srid(4326)
        return product



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import datetime
import numpy as np
from typing import Dict, Iterable
from geodataflow.pipeline.basictypes import AbstractWriter
from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.core.processingargs import ProcessingUtils


class TimeseriesPlot(AbstractWriter):
    """
    This module plots to image Time Series of values to visualize trends
    in counts or numerical values over time.
    """
    def __init__(self):
        AbstractWriter.__init__(self)
        self.connectionString = ''
        # Graph properties.
        self.title = 'Time series Plot'
        self.xLabel = 'Date'
        self.yLabel = 'Value'
        self.figureXSize = 800
        self.figureYSize = 600
        # Time Series properties.
        self.label = ''
        self.expressionValue = 'float(mean)'
        self.attributeDate = 'productDate'
        self.dateFormatter = '%Y-%m-%d'
        self.dateRange = 10

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'It plots to image Time Series of values to visualize trends in counts or numerical values over time.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Output'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'connectionString': {
                'description': 'Connection string or Filename of the image to output.',
                'dataType': 'string',
                'default': 'plot.png',
                'extensions': ['.png', '.jpg']
            },
            'title': {
                'description': 'Title of the Graph.',
                'dataType': 'string',
                'default': 'Time series Plot'
            },
            'xLabel': {
                'description': 'Label for the x-axis.',
                'dataType': 'string',
                'default': 'Date'
            },
            'yLabel': {
                'description': 'Label for the y-axis.',
                'dataType': 'string',
                'default': 'Value'
            },
            'figureXSize': {
                'description': 'The number of pixels of the image in the X direction.',
                'dataType': 'int',
                'default': 800
            },
            'figureYSize': {
                'description': 'The number of pixels of the image in the Y direction.',
                'dataType': 'int',
                'default': 600
            },
            'expressionValue': {
                'description':
                    'Algebraic expression to calculate the values, or list of them separated by commas.',
                'dataType': 'calc'
            },
            'attributeDate': {
                'description': 'Attribute containing the Date in format "%Y-%m-%d".',
                'dataType': 'string'
            },
            'label': {
                'description':
                    'Optional label, or list of them separated by commas, for the Legend. None by default.',
                'dataType': 'string'
            },
            'dateFormatter': {
                'description': 'Format pattern of Dates for the x-axis.',
                'dataType': 'string',
                'default': '%Y-%m-%d'
            },
            'dateRange': {
                'description': 'The interval between each iteration for the x-axis ticker.',
                'dataType': 'int',
                'default': 10
            }
        }

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns if this Module supports the specified ConnectionString and named StoreCapability.
        """
        file_name, file_ext = os.path.splitext(connection_string)
        return file_name and file_ext and file_ext in ['.png', '.jpg', '.jpeg']

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.geoext.commonutils import DataUtils

        for item_string in DataUtils.enumerate_single_connection_string(self.connectionString):
            #
            if not self.test_capability(item_string, StoreCapabilities.CREATE):
                raise Exception('Theimage  format "{}" is not supported!'.format(item_string))

            from geodataflow.core.schemadef import SchemaDef, GeometryType
            from geodataflow.geoext.dataset import DATASET_DEFAULT_SCHEMA_DEF

            schema_def = SchemaDef(type='RasterLayer',
                                   name=DataUtils.get_layer_name(item_string),
                                   srid=0,
                                   crs=None,
                                   geometryType=GeometryType.Polygon,
                                   envelope=[0, 0, self.figureXSize, self.figureYSize],
                                   fields=DATASET_DEFAULT_SCHEMA_DEF.copy())

            return schema_def

        return None

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        timeseries = []

        expressions = \
            self.expressionValue.split(',') if isinstance(self.expressionValue, str) else self.expressionValue
        labels = \
            self.label.split(',') if isinstance(self.label, str) and self.label else self.label

        if isinstance(expressions, list) and isinstance(labels, list) and len(expressions) != len(labels):
            raise Exception(
                'TimeseriesPlot does not support "expressionValue" and "label" settings with different size.')

        # Extract stream of Dates and array of Values.
        for feature in feature_store:
            date_ = feature.properties[self.attributeDate] if self.attributeDate else None
            values = [
                ProcessingUtils.eval_function(xpr, feature.properties, False)
                for xpr in expressions
            ]
            timeseries.append({'date': date_, 'values': values})

        # Refactoring stream of Date/Values to numpy arrays.
        timeseries = TimeseriesPlot._timeseries_np(timeseries)

        # Draw plot!
        if len(timeseries) > 0:
            import matplotlib
            import matplotlib.dates as m_dates
            matplotlib.use('agg')
            import matplotlib.pyplot as plt

            from geodataflow.geoext.gdalenv import GdalEnv
            from geodataflow.geoext.dataset import GdalDataset

            file_ext = os.path.splitext(self.connectionString)[1]
            relative_date = '%Y' not in self.dateFormatter
            has_labels = False

            # Main attributes of Plot.
            inv_dpi_ = 0.01
            fig, ax_ = plt.subplots(figsize=(inv_dpi_*self.figureXSize, inv_dpi_*self.figureYSize))
            if self.title:
                ax_.set_title(self.title)
            if self.xLabel:
                ax_.set_xlabel(self.xLabel)
            if self.yLabel:
                ax_.set_ylabel(self.yLabel)

            # Drawing Time Series collection.
            # TODO: Allow customization of symbology of Lines.
            settings = {}
            date_min = timeseries[0].min().astype(datetime.datetime)
            date_max = timeseries[0].max().astype(datetime.datetime)
            #
            for index, values in enumerate(timeseries[1:]):
                x = timeseries[0]
                y = values
                l_label = labels[index] if isinstance(labels, list) else None
                l_style = settings.get('linestyle', '-')
                l_width = settings.get('linewidth', 1)
                l_color = settings.get('color', None)
                m_style = settings.get('marker', None)
                m_wsize = settings.get('markersize', 3)

                ax_.plot(x, y,
                         color=l_color,
                         linestyle=l_style,
                         linewidth=l_width,
                         marker=m_style,
                         markersize=m_wsize,
                         label=l_label)

                has_labels = has_labels or l_label

            if relative_date:
                day_range = (date_max - date_min).days
                while day_range > 365:
                    day_range = day_range - 365
            else:
                day_range = (date_max - date_min).days

            ax_font_size = max(1, int(8 * self.figureXSize / 800))
            day_interval = max(1, 2 * int(day_range / self.dateRange))
            ax_.xaxis.set_major_locator(m_dates.DayLocator(interval=day_interval))
            ax_.xaxis.set_major_formatter(m_dates.DateFormatter(self.dateFormatter))

            for tick in ax_.xaxis.get_major_ticks():
                tick.label1.set_fontsize(max(1, ax_font_size - 1))

            fig.autofmt_xdate()
            if has_labels:
                ax_.legend(loc='best', fontsize=ax_font_size)

            ax_.grid()
            fig.savefig(self.connectionString, format=file_ext[1:])
            plt.close()

            # Load GDAL dataset from the rasterized Graph.
            gdal_env = GdalEnv.default()
            gdal = gdal_env.gdal()
            dataset = gdal.Open(self.connectionString, gdal.GA_ReadOnly)
            yield GdalDataset(dataset, gdal_env)

        pass

    @staticmethod
    def _timeseries_np(timeseries: Iterable[Dict]):
        """
        Convert stream of Date/Values to numpy arrays.
        """
        today_date = datetime.date.today()
        result = []

        for date_index, ts in enumerate(timeseries):
            date_ = ts.get('date')
            values = ts.get('values')

            if not date_:
                date_ = today_date + datetime.timedelta(days=date_index)
                date_ = date_.strftime('%Y-%m-%d')

            result.append([np.datetime64(date_)] + values)

        result = sorted(result, key=lambda x: x[0], reverse=False)
        result = np.transpose(np.array(result))
        return result



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import logging
from typing import Dict

from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.pipeline.basictypes import AbstractWriter


class RasterWriter(AbstractWriter):
    """
    Writes Datasets to a Geospatial RasterStore using GDAL providers.
    """
    def __init__(self):
        AbstractWriter.__init__(self)
        self.connectionString = ''
        self.formatOptions = []

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Writes Datasets to a Geospatial RasterStore using GDAL providers.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Output'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'connectionString': {
                'description':
                    'Connection string of the Raster Store (Common GDAL extensions are supported).',
                'dataType': 'string',
                'default': 'output.tif',
                'extensions': ['.tif', '.ecw', '.jp2', '.png', '.jpg']
            },
            'formatOptions': {
                'description': 'GDAL format options of output Dataset (Optional).',
                'dataType': 'string',
                'default': '-of COG',
            }
        }

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns this Module supports the specified ConnectionString and named StoreCapability.
        """
        from geodataflow.geoext.commonutils import GdalUtils
        return GdalUtils.test_gdal_capability(connection_string, capability)

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.geoext.commonutils import DataUtils

        for item_string in DataUtils.enumerate_single_connection_string(self.connectionString):
            #
            if not self.test_capability(item_string, StoreCapabilities.CREATE):
                raise Exception('The GDAL Driver of "{}" does not support Data creation!'.format(item_string))

            from geodataflow.core.schemadef import SchemaDef, GeometryType
            from geodataflow.geoext.dataset import DATASET_DEFAULT_SCHEMA_DEF

            schema_def = SchemaDef(type='RasterLayer',
                                   name=DataUtils.get_layer_name(item_string),
                                   srid=schema_def.crs.to_epsg(),
                                   crs=schema_def.crs,
                                   geometryType=GeometryType.Polygon,
                                   envelope=[1.e+20, 1.e+20, -1.e+20, -1.e+20],
                                   fields=DATASET_DEFAULT_SCHEMA_DEF.copy())

            return schema_def

        return None

    def run(self, dataset_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        dataset_count = 0

        from geodataflow.geoext.commonutils import DataUtils, GdalUtils
        from geodataflow.geoext.dataset import GdalDataset
        from geodataflow.geoext.gdalenv import GdalEnv

        gdal_env = GdalEnv.default()
        gdal = gdal_env.gdal()

        format_options = self.formatOptions or []
        format_options = format_options.split(' ') if isinstance(format_options, str) else format_options
        driver_options = [format_options[i + 1] for i, opt in enumerate(format_options) if opt == '-of']
        format_options = [format_options[i + 1] for i, opt in enumerate(format_options) if opt == '-co']

        connection_strings = list(DataUtils.enumerate_single_connection_string(self.connectionString))
        connection_index = 0
        connection_files = []

        for dataset in dataset_store:
            #
            if not isinstance(dataset, GdalDataset):
                raise Exception('RasterStore only accepts Datasets as input data.')

            if connection_index >= len(connection_strings):
                connection_string = connection_strings[0]
                layer_name = DataUtils.get_layer_name(connection_string)
                connection_string = \
                    DataUtils.replace_layer_name(connection_string, '{}_{}'.format(layer_name, connection_index))
            else:
                connection_string = connection_strings[connection_index]

            connection_files.append({
                'output_file': connection_string,
                'suffix': dataset.properties.get('productDate') or dataset.properties.get('suffix')
            })
            connection_index += 1

            driver = GdalUtils.get_gdal_driver(connection_string) \
                if not driver_options else gdal.GetDriverByName(driver_options[0])
            metadata = driver.GetMetadata()

            # Write GDAL Datasets.
            if gdal.DCAP_CREATECOPY in metadata and metadata[gdal.DCAP_CREATECOPY] == 'YES':
                output = \
                    driver.CreateCopy(connection_string, dataset.dataset(), strict=0, options=format_options)

                output.FlushCache()
                output = None
            else:
                info = dataset.get_metadata()
                cols = info.get('rasterSizeX')
                rows = info.get('rasterSizeY')
                band_count = info.get('rasterCount')
                data_type = info.get('dataType')
                no_data = info.get('noData')
                spatial_ref = GdalUtils.get_spatial_reference(info.get('srid'))
                envelope = dataset.get_envelope()
                pixel_size_x = info.get('pixelSizeX')
                pixel_size_y = info.get('pixelSizeY')

                # Create output GDAL Dataset.
                output = \
                    driver.Create(connection_string, cols, rows, band_count, data_type, format_options)

                gdal_dataset = dataset.dataset()
                output.SetGeoTransform([envelope[0], pixel_size_x, 0.0, envelope[3], 0.0, -pixel_size_y])
                output.SetProjection(spatial_ref.ExportToWkt())
                metadata = gdal_dataset.GetMetadata()
                if metadata is not None:
                    output.SetMetadata(metadata)

                for band_index in range(0, band_count):
                    band_s = gdal_dataset.GetRasterBand(band_index + 1)
                    raster = band_s.ReadAsArray(xoff=0, yoff=0, win_xsize=cols, win_ysize=rows)
                    band_s = None

                    band_t = output.GetRasterBand(band_index + 1)
                    band_t.Fill(no_data)
                    band_t.SetNoDataValue(no_data)
                    band_t.WriteArray(raster)
                    band_t.FlushCache()
                    band_t = None

                output.FlushCache()
                output = None

            dataset_count += 1
            yield dataset

        # Rename output files when we are getting a stream of outputs.
        if len(connection_files) > 1:
            layer_name = DataUtils.get_layer_name(connection_files[0]['output_file'])
            suffixes = {}

            for data_file in connection_files:
                output_file = data_file['output_file']
                suffix = data_file['suffix']

                if suffix and os.path.exists(output_file):
                    if suffixes.get(suffix) is not None:
                        suffixes[suffix] += 1
                        suffix = suffix + '_' + str(suffixes.get(suffix))
                    else:
                        suffixes[suffix] = 0

                    target_file = DataUtils.replace_layer_name(output_file, '{}_{}'.format(layer_name, suffix))
                    os.rename(output_file, target_file)

            pass

        logging.info('{:,} Datasets saved to "{}".'.format(dataset_count, connection_string))
        pass

    def finished_run(self, pipeline, processing_args) -> bool:
        """
        Finishing a Workflow on Geospatial data.
        """
        return True



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
from setuptools import find_packages, setup

BASEDIR = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))


# Metadata of GeodataFlow.
metadata = {}
with open(os.path.join(BASEDIR, 'geodataflow/__meta__.py'), 'r') as fp:
    exec(fp.read(), metadata)

with open(os.path.join(BASEDIR, 'README.md'), 'r') as fp:
    readme = fp.read()

with open(os.path.join(BASEDIR, 'requirements.txt'), 'r') as fp:
    all_reqs = fp.readlines()

install_requires = [x.strip()
                    for x in all_reqs if 'git+' not in x and '--' not in x and not x.startswith('#')]
dependency_links = [x.strip().replace('git+', '')
                    for x in all_reqs if 'git+' not in x and '--' not in x and not x.startswith('#')]

# Setup for GeodataFlow.
setup(
    name=metadata['__title__'],
    version=metadata['__version__'],
    description=metadata['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=metadata['__author__'],
    author_email=metadata['__email__'],
    url=metadata['__url__'],
    license=metadata['__license__'],
    packages=find_packages(exclude=['api', 'docs', 'tests', 'ui']),
    package_data={
        '': ['LICENSE', 'NOTICE'],
        'geodataflow': ['pipelineapp.default.settings']
    },
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
    python_requires='>=3.7',
    extras_require={
        'eodag': ['eodag>=2.4.0'],
        'gee': ['earthengine-api==0.1.320']
    },
    entry_points={
        'console_scripts': ['geodataflow = geodataflow.pipelineapp:pipeline_app']
    },
    project_urls={
        'Bug Tracker': 'https://github.com/ahuarte47/geodataflow/issues/',
        'Source Code': 'https://github.com/ahuarte47/geodataflow'
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import numpy as np
import logging
from typing import Dict, List
from geodataflow.pipeline.basictypes import AbstractFilter

# List of supported raster statistics.
ZONAL_STATS = [
    'min',
    'max',
    'mean',
    'count',
    'sum',
    'std',
    'median',
    'majority',
    'minority',
    'unique',
    'range',
    'nodataCount',
    'percentile_10',
    'percentile_25',
    'percentile_75',
    'percentile_90',
    'size'
]


def key_assoc_val(d, func, exclude=None):
    """
    Return the key associated with the value returned by func.
    """
    vs = list(d.values())
    ks = list(d.keys())
    key = ks[vs.index(func(vs))]
    return key


def get_percentile(stat):
    if not stat.startswith('percentile_'):
        raise ValueError("must start with 'percentile_'")

    qstr = stat.replace("percentile_", '')
    q = float(qstr)

    if q > 100.0:
        raise ValueError('percentiles must be <= 100')
    if q < 0.0:
        raise ValueError('percentiles must be >= 0')

    return q


class RasterStats(AbstractFilter):
    """
    The Filter summarizes geospatial raster datasets and transform them to vector geometries.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.stats = ['median']
        self.bandIndex = 0
        self.polygonize = False

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Summarizes geospatial raster datasets and transform them to vector geometries.'

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Stats'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Raster'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        zonal_stats = sorted(ZONAL_STATS)

        return {
            'stats': {
                'description':
                    'Method, or list of methods separated by commas, ' +
                    'of summarizing and aggregating the raster values of input Datasets--.',
                'dataType': 'string',
                'default': 'median',
                'options': zonal_stats,
                'labels': zonal_stats
            },
            "bandIndex": {
                'description': 'Index of Band from which to calculate Raster statistics.',
                'dataType': 'int',
                'default': 0
            },
            "polygonize": {
                'description':
                    'Creates vector polygons for all connected regions of no-data pixels in the raster.',
                'dataType': 'bool',
                'default': False
            }
        }

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.core.schemadef import GeometryType, DataType, FieldDef

        stat_names = self.stats if isinstance(self.stats, list) else str(self.stats).split(',')
        for stat_name in stat_names:
            if stat_name not in ZONAL_STATS:
                raise Exception('The raster statistic "{}" is not supported by RasterStats.'.format(stat_name))

        schema_def = schema_def.clone()
        schema_def.type = 'FeatureLayer'
        schema_def.geometryType = GeometryType.Polygon
        schema_def.fields += [
            FieldDef(name=stat_name, data_type=DataType.Float) for stat_name in stat_names
        ]
        return schema_def

    def run(self, data_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        from geodataflow.geoext.dataset import GdalDataset

        stat_names = self.stats if isinstance(self.stats, list) else str(self.stats).split(',')
        band_index = self.bandIndex
        feature_index = 0

        for dataset in data_store:
            if not isinstance(dataset, GdalDataset):
                raise Exception('RasterStats only accepts Datasets as input data.')

            gdal_dataset = dataset.dataset()
            raster_size_x = gdal_dataset.RasterXSize
            raster_size_y = gdal_dataset.RasterYSize
            properties = dataset.properties.copy()

            # Extract or Polygonize Geometry.
            geometry = dataset.polygonize(band_index) if self.polygonize else dataset.geometry
            if geometry is None:
                logging.warning('Polygonize returns an empty Geometry, input Dataset is skipped.')
                continue

            # Calculate Zonal statistics.
            rs_band = gdal_dataset.GetRasterBand(band_index + 1)
            raster = rs_band.ReadAsArray(xoff=0, yoff=0, win_xsize=raster_size_x, win_ysize=raster_size_y)
            nodata = rs_band.GetNoDataValue()
            properties.update(RasterStats.zonal_statistics(raster, nodata, stat_names))
            raster = None
            rs_band = None

            feature = type('Feature', (object,), {
                'type': 'Feature',
                'fid': feature_index,
                'properties': properties,
                'geometry': geometry,
            })
            feature_index += 1
            yield feature

        pass

    @staticmethod  # noqa: C901
    def zonal_statistics(raster, no_data: float, stats: List[str]):  # noqa: C901
        """
        Zonal statistics of raster values aggregated to vector geometries.
        """
        raster = raster.astype(float)
        no_data = float(no_data)
        feature_stats = {}

        # Run the counter once, only if needed.
        run_count = 'majority' in stats or 'minority' in stats or 'unique' in stats

        # Set nodata mask.
        is_no_data = (raster == no_data) | np.isnan(raster)
        # Mask the source data array.
        masked = np.ma.MaskedArray(raster, mask=is_no_data)

        # Fill stats.
        if masked.compressed().size == 0:
            # nothing here, fill with None and move on.
            feature_stats = dict([(stat, None) for stat in stats])
            # special case, zero makes sense here.
            if 'count' in stats:
                feature_stats['count'] = 0
        else:
            if run_count:
                keys, counts = np.unique(masked.compressed(), return_counts=True)
                try:
                    pixel_count = dict(zip([k.item() for k in keys], [c.item() for c in counts]))
                except AttributeError:
                    pixel_count = dict(zip([np.asscalar(k) for k in keys], [np.asscalar(c) for c in counts]))
            else:
                pixel_count = {}

            if 'min' in stats:
                feature_stats['min'] = float(masked.min())
            if 'max' in stats:
                feature_stats['max'] = float(masked.max())
            if 'mean' in stats:
                feature_stats['mean'] = float(masked.mean())
            if 'count' in stats:
                feature_stats['count'] = int(masked.count())
            if 'sum' in stats:
                feature_stats['sum'] = float(masked.sum())
            if 'std' in stats:
                feature_stats['std'] = float(masked.std())
            if 'median' in stats:
                feature_stats['median'] = float(np.median(masked.compressed()))
            if 'majority' in stats:
                feature_stats['majority'] = float(key_assoc_val(pixel_count, max))
            if 'minority' in stats:
                feature_stats['minority'] = float(key_assoc_val(pixel_count, min))
            if 'unique' in stats:
                feature_stats['unique'] = len(list(pixel_count.keys()))
            if 'range' in stats:
                rmin = float(masked.min())
                rmax = float(masked.max())
                feature_stats['range'] = rmax - rmin

            for pctile in [s for s in stats if s.startswith('percentile_')]:
                q = get_percentile(pctile)
                pctarr = masked.compressed()
                feature_stats[pctile] = np.percentile(pctarr, q)

        if 'nodataCount' in stats:
            feature_stats['nodataCount'] = int(is_no_data.sum())
        if 'size' in stats:
            feature_stats['size'] = raster.size

        return feature_stats



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from typing import Dict
from geodataflow.pipeline.basictypes import AbstractFilter


class RasterClip(AbstractFilter):
    """
    The Filter clips input Rasters by a Geometry.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.clipGeometries = ''
        self.cutline = True
        self.allTouched = True

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Clip'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Clips input Rasters by a Geometry.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Raster'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'clipGeometries': {
                'description': 'Collection of Geometries that will clip input Features.',
                'dataType': 'input'
            },
            'cutline': {
                'description': 'Clipping using lines of Geometries, otherwise just envelopes are used. True by default.',
                'dataType': 'bool',
                'default': True
            },
            'allTouched': {
                'description':
                    'Ensures that that all pixels overlapping the cutline polygon will be selected, ' +
                    'not just those whose center point falls within the polygon. True by default.',
                'dataType': 'bool',
                'default': True
            }
        }

    def run(self, data_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        from geodataflow.geoext.dataset import GdalDataset

        clipping_geoms = [
            obj.geometry for obj in self.enumerate_inputs(self.clipGeometries)
        ]

        if clipping_geoms:
            from geodataflow.geoext.commonutils import GeometryUtils

            schema_def = self.pipeline_args.schema_def
            if schema_def.srid:
                schema_crs = GeometryUtils.get_spatial_crs(schema_def.srid)

                for g_index, g in enumerate(clipping_geoms):
                    if g.get_srid():
                        clipping_crs = GeometryUtils.get_spatial_crs(g.get_srid())
                        transform_fn = GeometryUtils.create_transform_function(clipping_crs, schema_crs)
                        clipping_geoms[g_index] = transform_fn(g)

            for dataset in data_store:
                if not isinstance(dataset, GdalDataset):
                    raise Exception('RasterClip only accepts Datasets as input data.')

                for g in clipping_geoms:
                    if g.intersects(dataset.geometry):
                        new_dataset = dataset.warp(
                            output_crs=None, output_geom=g, cutline=self.cutline, all_touched=self.allTouched
                        )
                        dataset.recycle()
                        dataset = new_dataset

                yield dataset
        else:
            for dataset in data_store:
                yield dataset

        pass
